//Polygon.cpp
#include <assert.h>
#include <d3d9.h>
#include <d3dx9.h>
#include "DxPolygon.h"
#include "MeshObj.h"

namespace DataType
{
	int Type = ( D3DFVF_XYZ | D3DFVF_NORMAL | D3DFVF_DIFFUSE | D3DFVF_TEX1 );
};


namespace DataType
{
	int WireType = ( D3DFVF_XYZ | D3DFVF_DIFFUSE );
};

void ComputeNormal( D3DXVECTOR3* out, D3DXVECTOR3* p0, D3DXVECTOR3* p1, D3DXVECTOR3* p2 )
{
	D3DXVECTOR3 u = *p1 - *p0;  
	D3DXVECTOR3 v = *p2 - *p0;

  	D3DXVec3Cross(out, &u, &v);
  	D3DXVec3Normalize(out, out);

/*
	D3DXVECTOR3 out;
	D3DXVECTOR3 p0 = D3DXVECTOR3( -1, -1, 0 );
	D3DXVECTOR3 p1 = D3DXVECTOR3( -1,  1, 0 );
	D3DXVECTOR3 p2 = D3DXVECTOR3(  1,  0, 0 );

	ComputeNormal( &out, &p0, &p1, &p2 ); //0, 0, -1 --> ޼ ǥ迡 ð Ѵ.
	ComputeNormal( &out, &p0, &p2, &p1 ); //0, 0, 1 -->  ǥ迡 ݽð Ѵ.
*/
}

void ComputePolygonNormal( DxPolygon* polygon )
{
#if  1
	PointArray& vrtArr = polygon->GetPosition();
	int nSize = (int)vrtArr.size();
	assert( nSize >= 3 );

	D3DXVECTOR3 out;
	for( int i = 0; i < nSize; i += 3 )
	{
		D3DXVECTOR3 v0( vrtArr[i].x, vrtArr[i].y, vrtArr[i].z );
		D3DXVECTOR3 v1( vrtArr[i+1].x, vrtArr[i+1].y, vrtArr[i+1].z );
		D3DXVECTOR3 v2( vrtArr[i+2].x, vrtArr[i+2].y, vrtArr[i+2].z );
		ComputeNormal( &out, &v0, &v1, &v2 );
		polygon->PushBackNormal( DataType::Point3( out.x, out.y, out.z )  );
		polygon->PushBackNormal( DataType::Point3( out.x, out.y, out.z )  );
		polygon->PushBackNormal( DataType::Point3( out.x, out.y, out.z )  );
	}
#else
	typedef std::vector<int> IndexArrayComponent;

	PointArray& vrtArr = polygon->GetPosition();
	int nSize = (int)vrtArr.size();
	if( nSize < 3 )
		return;

	//assert( nSize >= 3 );
	D3DXVECTOR3* normal = new D3DXVECTOR3[ nSize ];
	memset( normal, 0, nSize * sizeof( D3DXVECTOR3 ) );

	IndexArrayComponent* sharedNormalIndex = new IndexArrayComponent[ nSize ];
	for( int i = 0; i < nSize; ++i )
	{
		for( int k = 0; k < nSize; ++k )
		{
			if( i != k )
			{
				if( vrtArr[i].x == vrtArr[k].x && vrtArr[i].y == vrtArr[k].y && vrtArr[i].z == vrtArr[k].z )
				{
					sharedNormalIndex[i].push_back( k );
				}
			}
		}	
	}

	D3DXVECTOR3 out;
	for( int i = 0; i < nSize; i += 3 )
	{
		D3DXVECTOR3 v0( vrtArr[i].x, vrtArr[i].y, vrtArr[i].z );
		D3DXVECTOR3 v1( vrtArr[i+1].x, vrtArr[i+1].y, vrtArr[i+1].z );
		D3DXVECTOR3 v2( vrtArr[i+2].x, vrtArr[i+2].y, vrtArr[i+2].z );
		ComputeNormal( &out, &v0, &v1, &v2 );

		normal[i+0] += out;
		normal[i+1] += out;
		normal[i+2] += out;

		int nCount = (int)sharedNormalIndex[i].size();
		IndexArrayComponent& indexArr0 = sharedNormalIndex[i];
		for( int k = 0; k < nCount; ++k )
		{
			int index = indexArr0[k];
			normal[ index ] += out;
		}

		nCount = (int)sharedNormalIndex[i+1].size();
		IndexArrayComponent& indexArr1 = sharedNormalIndex[i+1];
		for( int k = 0; k < nCount; ++k )
		{
			int index = indexArr1[k];
			normal[ index ] += out;
		}

		nCount = (int)sharedNormalIndex[i+2].size();
		IndexArrayComponent& indexArr2 = sharedNormalIndex[i+2];
		for( int k = 0; k < nCount; ++k )
		{
			int index = indexArr2[k];
			normal[ index ] += out;
		}
	}

	for( int i = 0; i < nSize; ++i )
	{
		D3DXVec3Normalize(&normal[i], &normal[i]);
		polygon->PushBackNormal( DataType::Point3( normal[i].x, normal[i].y, normal[i].z )  );
	}

	delete [] normal;
	delete [] sharedNormalIndex;
#endif
}

DxPolygon::DxPolygon()
{
	m_translate = m_rotate = D3DXVECTOR3( 0, 0, 0 );
	m_scale = D3DXVECTOR3( 1, 1, 1 );

	m_pVertex = NULL;
	m_pWireVertex = NULL;
	m_nCount = 0;
	m_color =  D3DCOLOR_XRGB(255, 255, 255);
	m_pTexture = NULL;
	m_bWireFrame = false;
	m_bLighting = TRUE;
}
//------------------------------------------------------------------------------------------------
DxPolygon::~DxPolygon()
{
	Cleanup();
}
//------------------------------------------------------------------------------------------------
void DxPolygon::Cleanup()
{
	if( m_pVertex )
	{
		delete [] m_pVertex;
		m_pVertex = NULL;
		delete [] m_pWireVertex;
		m_pWireVertex = NULL;
		m_nCount = 0;

		_RELEASE_<LPDIRECT3DTEXTURE9>( m_pTexture );
	}
}
//------------------------------------------------------------------------------------------------
void DxPolygon::LoadObj( const char* szName )
{
	mesh* object = new mesh();
	if( object->open( szName ) == false )
	{
		delete object;
		return;
	}

	std::vector<mesh::Vec3>& vList = object->vList;
	std::vector<mesh::Vec3>& nList = object->nList;
	std::vector<mesh::Vec2>& tList = object->tList;
	std::vector<mesh::FACE>& faceList = object->faceList;
	int n = (int) faceList.size();

	for( int i = 0; i < n; ++i )
	{
		mesh::Vertex v0 = faceList[i].v[0];
		mesh::Vertex v1 = faceList[i].v[1];
		mesh::Vertex v2 = faceList[i].v[2];

		mesh::Vec3 vec = vList[ v0.v ]; mesh::Vec3 nom = nList[ v0.n ];
		m_position.push_back( DataType::Point3( vec[0], vec[1], vec[2] ) );
		m_normal.push_back( DataType::Point3( nom[0], nom[1], nom[2] ) );

		vec = vList[ v1.v ];  nom = nList[ v1.n ];
		m_position.push_back( DataType::Point3( vec[0], vec[1], vec[2] ) );
		m_normal.push_back( DataType::Point3( nom[0], nom[1], nom[2] ) );

		vec = vList[ v2.v ];  nom = nList[ v2.n ];
		m_position.push_back( DataType::Point3( vec[0], vec[1], vec[2] ) );
		m_normal.push_back( DataType::Point3( nom[0], nom[1], nom[2] ) );
	}

	if( tList.empty() == false )
	{
		for( int i = 0; i < n; ++i )
		{
			mesh::Vertex v0 = faceList[i].v[0];
			mesh::Vertex v1 = faceList[i].v[1];
			mesh::Vertex v2 = faceList[i].v[2];

			mesh::Vec2 uv = tList[ v0.t ];
			m_texUV.push_back( DataType::Point2( uv[0], uv[1] ) );

			uv = tList[ v1.t ];
			m_texUV.push_back( DataType::Point2( uv[0], uv[1] ) );

			uv = tList[ v2.t ];
			m_texUV.push_back( DataType::Point2( uv[0], uv[1] ) );
		}
	}
	delete object;
}
//------------------------------------------------------------------------------------------------
bool DxPolygon::SetTexture( const LPDIRECT3DDEVICE9 pDevice, const char* szFileName )
{
	HRESULT hr = D3DXCreateTextureFromFile( pDevice, szFileName, &m_pTexture );
	return SUCCEEDED( hr );
}
//------------------------------------------------------------------------------------------------
void DxPolygon::SetVertex( DataType::Vertex* pVertex, int nNum )
{
	if( m_nCount > 0 && m_nCount != nNum )
		Cleanup();

	m_nCount = nNum;
	if( m_pVertex == NULL )
	{
		m_pVertex = new DataType::Vertex [ nNum ];
		m_pWireVertex = new DataType::WireVertex [ nNum ];
	}

	memcpy( m_pVertex, pVertex, nNum * sizeof( DataType::Vertex ) );
	for( int k = 0; k < nNum; ++k )
	{
		m_pWireVertex[k].x = pVertex[k].x;
		m_pWireVertex[k].y = pVertex[k].y;
		m_pWireVertex[k].z = pVertex[k].z;
		m_pWireVertex[k].color = 0xff0000;
	}
}
//------------------------------------------------------------------------------------------------
void DxPolygon::PushBackPosition( DataType::Point3& pos )
{
	m_position.push_back( pos );
}
//------------------------------------------------------------------------------------------------
void DxPolygon::PushBackNormal( DataType::Point3& pos )
{
	m_normal.push_back( pos );
}
//------------------------------------------------------------------------------------------------
void DxPolygon::ReComputeNormal()
{
	if( m_pVertex )
		Cleanup();

	m_normal.clear();
	ComputePolygonNormal( this );
}
//------------------------------------------------------------------------------------------------
void DxPolygon::CopyData()
{
	int nNum = m_nCount = (int)m_position.size();
	int nNormalNum = (int)m_normal.size();
	int nUVNum = (int)m_texUV.size();

	m_pVertex = new DataType::Vertex [ nNum ];
	m_pWireVertex = new DataType::WireVertex [ nNum ];

	for( int i = 0; i < nNum; ++i )
	{
		m_pVertex[i].x = m_position[i].x;
		m_pVertex[i].y = m_position[i].y;
		m_pVertex[i].z = m_position[i].z;
		
		if( nNum == nNormalNum )
		{
			m_pVertex[i].nx = m_normal[i].x;
			m_pVertex[i].ny = m_normal[i].y;
			m_pVertex[i].nz = m_normal[i].z;
		}
		else
		{
			//TODO:  ڵ  ߰
			m_pVertex[i].nx = 1.0f;
			m_pVertex[i].ny = 1.0f;
			m_pVertex[i].nz = 1.0f;
		}

		if( nNum == nUVNum )
		{
			m_pVertex[i].u = m_texUV[i].x;
			m_pVertex[i].v = m_texUV[i].y;
		}
		else
		{
			m_pVertex[i].u = 0;
			m_pVertex[i].v = 0;			
		}

		m_pVertex[i].color = m_color;

		m_pWireVertex[i].x = m_position[i].x;
		m_pWireVertex[i].y = m_position[i].y;
		m_pWireVertex[i].z = m_position[i].z;
		m_pWireVertex[ i ].color = 0xff0000;
	}
}
//------------------------------------------------------------------------------------------------
void DxPolygon::Render( const LPDIRECT3DDEVICE9 device )
{
	if( m_pVertex == NULL )
	{
		if( m_position.empty() )
			return;

		CopyData();
	}

	//  ̵
	D3DXMATRIX	matRot, matTrans, matScale;
	D3DXMatrixTranslation( &matTrans, m_translate.x, m_translate.y, m_translate.z );
	D3DXMatrixRotationYawPitchRoll( &matRot, m_rotate.y, m_rotate.x, m_rotate.z );
	D3DXMatrixScaling( &matScale, m_scale.x, m_scale.y, m_scale.z );
	D3DXMATRIXA16  matWorld = matScale * matRot * matTrans;
	device->SetTransform( D3DTS_WORLD, &matWorld );

	HRESULT hr;
	if( m_pTexture )
		device->SetTexture( 0, m_pTexture );
	// Draw the vertex list
	device->SetFVF( DataType::Type );
	device->SetRenderState( D3DRS_LIGHTING, m_bLighting );
	device->SetRenderState( D3DRS_FILLMODE, D3DFILL_SOLID );
	hr = device->DrawPrimitiveUP( D3DPT_TRIANGLELIST, m_nCount/3, m_pVertex, sizeof(DataType::Vertex ) );

	//Wireframe ׸ ʴ´.
	if( m_bWireFrame )
	{
		device->SetFVF( DataType::WireType );
		device->SetRenderState( D3DRS_LIGHTING, FALSE );
		device->SetRenderState( D3DRS_FILLMODE, D3DFILL_WIREFRAME );
		hr = device->DrawPrimitiveUP( D3DPT_TRIANGLELIST, m_nCount/3, m_pWireVertex, sizeof(DataType::WireVertex ) );
	}

	if( m_pTexture )
		device->SetTexture( 0, NULL );
}
//-------------------------------------------------------------------------------------------------
const D3DXVECTOR3& DxPolygon::GetTranslate()
{
	return m_translate;
}
//-------------------------------------------------------------------------------------------------
const D3DXVECTOR3& DxPolygon::GetRotate()
{
	return m_rotate;
}
//-------------------------------------------------------------------------------------------------
const D3DXVECTOR3& DxPolygon::GetScale()
{
	return m_scale;
}
//-------------------------------------------------------------------------------------------------
void DxPolygon::SetTranslate( float x, float y, float z )
{
	m_translate = D3DXVECTOR3( x, y, z );
}
//-------------------------------------------------------------------------------------------------
void DxPolygon::SetRotate( float x, float y, float z )
{
	m_rotate = D3DXVECTOR3( x, y, z );
}
//-------------------------------------------------------------------------------------------------
void DxPolygon::SetScale( float x, float y, float z )
{
	m_scale = D3DXVECTOR3( x, y, z );
}
//-------------------------------------------------------------------------------------------------
void DxPolygon::OnLighting( bool bFlag )
{
	if( bFlag )
		m_bLighting = TRUE;
	else
		m_bLighting = FALSE;
}
//-------------------------------------------------------------------------------------------------
void DxPolygon::OnWireFrame( bool bFlag )
{
	m_bWireFrame = bFlag;
}